package com.wudsn.productions.atari800.aspongy;

import java.awt.FlowLayout;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.util.zip.Deflater;

import javax.swing.JFrame;

public final class ASpongyGenerator {

    private static final String FOLDER_PATH = "C:\\jac\\system\\Atari800\\Programming\\Demos\\ASpongy";
    private static final String FILE_NAME = "ASpongy-Animation.bin";

    public static final int ROTATION_SPEED1 = 1;
    public static final int ROTATION_SPEED2 = 1;

    private JFrame frame;
    private List<ASpongy> spongies;

    private Random random;
    private int maxColor;
    private int minColor;

    public static void main(String[] args) {

	ASpongyGenerator instance = new ASpongyGenerator();
	try {
	    instance.run();
	} catch (IOException ex) {
	    ex.printStackTrace();
	    System.exit(1);
	}
    }

    private ASpongyGenerator() {

	frame = new JFrame("ASpongy");
	frame.addWindowListener(new WindowAdapter() {
	    @Override
	    public void windowClosing(WindowEvent e) {
		System.exit(0);
	    }
	});

	random = new Random();

    }

    private void run() throws IOException {
	int[] palette256;
	int[] palette16;
	int[] palette2;
	palette256 = new int[256];
	palette16 = new int[256];
	palette2 = new int[2];

	for (int i = 0; i < 256; i++) {
	    int color = 255 - i;
	    color = color >>> 4;

	    color = color << 4;
	    color = color << 16 | color << 8 | color;
	    palette16[i] = color;

	    color = 255 - i;
	    color = color << 16 | color << 8 | color;
	    palette256[i] = color;

	}
	palette2[0] = 0x000000;
	palette2[1] = 0xffffff;

	ASpongy instance;
	spongies = new ArrayList<ASpongy>();
	instance = new ASpongyDouble();
	instance.setPalette(palette256);
	spongies.add(instance);
	instance = new ASpongyFloat();
	instance.setPalette(palette256);
	spongies.add(instance);
	// instance = new JSpongy3();
	// instance.setPalette(palette256);
	// spongies.add(instance);
	instance = new ASpongy5Short();
	instance.setPalette(palette256);
	spongies.add(instance);

	ASpongy instance16;
	instance16 = new ASpongy5Short();
	instance16.setPalette(palette16);
	spongies.add(instance16);

	ASpongy instance2;
	instance2 = new ASpongy5Short();
	instance2.setPalette(palette2);
	spongies.add(instance2);

	frame.setLayout(new FlowLayout());
	for (ASpongy spongy : spongies) {
	    frame.add(spongy.getContainer());
	    spongy.init();
	}
	frame.setVisible(true);
	frame.pack();

	FileOutputStream fos = null;

	File file = new File(FOLDER_PATH, FILE_NAME);
	fos = new FileOutputStream(file);
	try {
	    renderLoop(file, fos, instance16);
	} finally {
	    fos.close();
	}

    }

    private void renderLoop(File file, OutputStream fos, ASpongy atariInstance) throws IOException {
	int t = 0;
	int tmax = 128;
	System.out.println("Maximum Time:" + tmax);

	boolean saveAtari = true;

	int[] imageData = new int[ASpongy.XSIZE * ASpongy.YSIZE];

	int screenSize = 2048;
	int bankSize = 16384;
	int screensPerBank = bankSize / screenSize;
	byte[] screenData = new byte[screenSize];
	int screenNumber = 0;
	int bankNumber = 0;
	byte[] bankData = new byte[bankSize];

	while (true) {
	    float angleX, angleY;
	    angleX = (float) (2.0f * Math.PI * t * ROTATION_SPEED1) / 128;

	    angleY = (float) (2.0f * Math.PI * t * ROTATION_SPEED2) / 128;

	    long minDuration = Long.MAX_VALUE;
	    ASpongy minSpongy = null;
	    for (ASpongy spongy : spongies) {
		long startTimeNanos = System.nanoTime();
		for (int i = 0; i < imageData.length; i++) {
		    imageData[i] = 0;
		}
		try {
		    spongy.frame(angleX, angleY, t, imageData);
		} catch (Exception ex) {
		    ex.printStackTrace();
		}
		long duration = System.nanoTime() - startTimeNanos;
		if (duration < minDuration) {
		    minDuration = duration;
		    minSpongy = spongy;
		}
		boolean atari = (spongy == atariInstance);
		spongy.paint(imageData, duration, atari);
	    }
	    if (minSpongy != null) {
		minSpongy.setCurrentMinimumIndicator();
	    }
	    try {
		Thread.sleep(1000 / 60);

	    } catch (InterruptedException ex) {
		throw new RuntimeException(ex);
	    }
	    if (saveAtari) {

		int screenOffset = 0;
		int AXSIZE = 80;
		int AYSIZE = 51;
		int pxstart = (ASpongy.XSIZE - AXSIZE) / 2;
		int pystart = (ASpongy.YSIZE - AYSIZE) / 2;

		maxColor = 0;
		minColor = 255;
		for (int py = 0; py < AYSIZE; py++) {
		    for (int px = 0; px < AXSIZE; px = px + 2) {
			int imageOffset = (pxstart + px) + (pystart + py) * ASpongy.XSIZE;
			int color1 = imageData[imageOffset++];
			int color2 = imageData[imageOffset];
			byte screenByte = (byte) (16 * dither(color1) + dither(color2));
			screenData[screenOffset++] = screenByte;
		    }
		}
		System.out.println("screenNumner=" + screenNumber + " maxColor=" + maxColor + " minColor=" + minColor);

		System.arraycopy(screenData, 0, bankData, screenNumber * screenSize, screenData.length);
		screenNumber++;
		if (screenNumber == screensPerBank) {

//		    for (int i = 0; i < bankData.length; i++) {
//			bankData[i] = (byte) i;
//		    }

		    // Compress the bytes
		    Deflater deflater = new Deflater(Deflater.BEST_COMPRESSION);
		    deflater.setInput(bankData);
		    deflater.finish();
		    byte[] compressedData = new byte[bankData.length];
		    int compressedDataLength = deflater.deflate(compressedData);
		    fos.write(255);
		    fos.write(255);
		    int screenStart = 0x8000;
		    int screenEnd = screenStart + compressedDataLength - 1 - 2;
		    fos.write(screenStart & 0xff);
		    fos.write(screenStart >> 8);
		    fos.write(screenEnd & 0xff);
		    fos.write(screenEnd >> 8);

		    fos.write(compressedData, 2, compressedDataLength - 2);
		    System.out.println("Compressed " + bankSize + " bytes to " + compressedDataLength + " bytes.");
		    // Write INIADR segment to jump to $2000 to set next
		    // extended RAM bank and deflate loaded data.
		    fos.write(0xff);
		    fos.write(0xff);
		    fos.write(0xe2);
		    fos.write(0x02);
		    fos.write(0xe3);
		    fos.write(0x02);
		    fos.write(0x00);
		    fos.write(0x20);

		    screenNumber = 0;
		    bankNumber++;
		}
	    }
	    t++;

	    if (t >= tmax) {
		t = 0;
		if (saveAtari) {
		    fos.close();
		    atariInstance.printVariables();
		    System.out.println("Output file '" + file.getAbsolutePath() + "' closed after " + bankNumber
			    + "banks with " + file.length() + " bytes.");
		    saveAtari = false;
		}

	    }
	}
    }

    private int dither(int color) {

	color = (color * 256) / 225;
	if (color > maxColor) {
	    maxColor = color;
	}
	if (color < minColor) {
	    minColor = color;
	}
	// color = 255-color;
	color += -5 + random.nextInt(10);

	if (color < 0) {
	    color = 0;
	}
	if (color > 255) {
	    color = 255;
	}
	color = color / 16;

	return color;
    }
}